package edu.sdsc.sirius.surface;

import java.util.*;

import edu.sdsc.mbt.*;
import edu.sdsc.mbt.util.*;
import edu.sdsc.sirius.util.*;
import edu.sdsc.mbt.viewables.*;

/**
 * This class encapsulates data necessary to generate a contiguous patch of surface for a given
 * set of atoms. The dot surface is created based on the Connolly algorithm, and dot coordinates
 * with the associated normal vectors are stored. Each vertex is assigned to its parent atom
 * for further manipulation, such as visibility and coloring
 * @author Sasha Buzko
 *
 */
public class SurfaceComponent extends StructureComponent {
//  Add for internationalization
	Locale locale;
	private ResourceBundle reslang;
	reslang = ResourceBundle.getBundle("SiriusLangRes", locale);
	private String[] _M_Symbol = new String[10];
// end		
	private Vector atoms;
	private double probeRadius = 1.4;
	
	private int quality = StylesPreferences.SURFACE_QUALITY_LOW;
	
	public boolean visible = true;
	public boolean selected = false;
	
	private HashMap points = new HashMap();//Atom -> Vector of double[6] coordinate points + normal vector coordinates
	
	/**
	 * Generate a SurfaceComponent with dots and associated normal vectors
	 * @param atoms - set of atoms to be surfaced
	 * @param structure - structure the atoms belong to
	 * @param quality - level of quality
	 * @param isolated - if true, neighboring atoms outside of the surfaced set are ignored for clash checking
	 */
	public SurfaceComponent(Vector atoms, Structure structure, int quality, boolean isolated){
		
		this.atoms = atoms;
		this.quality = quality;
		this.structure = structure;
		
		//generate vectors to be used for probe placement on each atom
		//since the vectors will need to sample every direction independent of the specific atomic coordinates,
		//they may well be pre-generated to avoid unnecessary overhead
		//hydrogen vectors need to be generated separately due to the big difference in its size compared to the heavy atoms
		
		double spacing = 0.8;
		if (quality == StylesPreferences.SURFACE_QUALITY_MEDIUM){
			spacing = 0.5;
		}
		else if (quality == StylesPreferences.SURFACE_QUALITY_HIGH){
			spacing = 0.25;
		}
		else if (quality == StylesPreferences.SURFACE_QUALITY_ULTRA){
			spacing = 0.1;
		}
		
		Vector vectors = computeVectors(1.7, spacing);
		Vector hVectors = computeVectors(1.2, spacing);
		
		
//		System.out.println("vectors = " + vectors.size());
		
		//generate pairs of atoms within a reasonable distance from each other
		//pairs are generated for the entire structure - just in case this is a subset surface
		Vector pairs = new Vector();
		
		Vector doubles = null;
		
		if (isolated){
			doubles = BondFactory.generateCovalentBonds(atoms, probeRadius * 2 + 2.5 * 2, false);
		}
		else{
			doubles = BondFactory.generateCovalentBonds(structure.getStructureMap().getAtoms(), probeRadius * 2 + 2.5 * 2, false);
		}
		//the pairs Vector contains Bond objects, but they actually represent closely located pairs of atoms
		HashMap atomMap = new HashMap();//Atom -> Vector of neighbor Atoms
		for (int i = 0; i < doubles.size(); i++){
			Bond b = (Bond)doubles.get(i);
			Atom a0 = b.getAtom(0);
			Atom a1 = b.getAtom(1);
			
			if (a0.compound.equals("HOH") || a0.compound.equals("WAT")) continue;
			if (a1.compound.equals("HOH") || a1.compound.equals("WAT")) continue;
			
			AtomPair pair = new AtomPair(a0, a1);
			pairs.add(pair);
			
			Vector aa0 = null;
			if (atomMap.containsKey(a0)){
				aa0 = (Vector)atomMap.get(a0);
			}
			else{
				aa0 = new Vector();
				atomMap.put(a0, aa0);
			}
			
			aa0.add(a1);
			
			Vector aa1 = null;
			if (atomMap.containsKey(a1)){
				aa1 = (Vector)atomMap.get(a1);
			}
			else{
				aa1 = new Vector();
				atomMap.put(a1, aa1);
			}
			
			aa1.add(a0);
			
		}
		
		int total = atoms.size() + pairs.size()/50;//total number of operations to perform
		int onePercent = total/100;
		float percentDone = 0;
		
// 		Status.progress(0.0f, "Creating surface...");
		Status.progress(0.0f, M_Creatingsurface);
		
//		System.out.println("atom count = " + atoms.size());

		//walk through the atoms and place the probe around the VdW sphere to get the contact points
		//at each position, check whether the probe intersects any of the other atoms
		//this generates points for all convex fragments
		int counter = 0;
		for (int i = 0; i < atoms.size(); i++){
			Atom a = (Atom)atoms.get(i);
			if (a.compound.equals("HOH") || a.compound.equals("WAT")) continue;
			
			double R = ElementProperties.getAtomRadius(a.element);
			double[] center = a.coordinate;
			
			double distance = R + probeRadius;//distance between atom center and probe center
			
			Vector pointList = new Vector();
			
			//walk through the set of precomputed vectors to get the center of the probe
			Vector set = vectors;
			if (a.element == 1){
				set = hVectors;
			}
			
			Vector neighbors = (Vector)atomMap.get(a);
			for (int j = 0; j < set.size(); j++){
				double[] v = (double[])set.get(j);
				
				double[] probe = StructureMath.getPointOnVector(center, v, distance);
				
				//check whether probe (center) is within a clashing distance with any other atoms
				boolean clash = false;
				if (neighbors != null && neighbors.size() != 0){
					for (int k = 0 ; k < neighbors.size(); k++){
						Atom aa = (Atom)neighbors.get(k);
						if (aa == a) continue;
						
						if (Algebra.distance(probe, aa.coordinate) < (probeRadius + ElementProperties.getAtomRadius(aa.element))){
							clash = true;
							break;
						}
					}
					
					if (clash) continue;//go on to the next vector - this one produced a probe position with a clash
				}
				
				//no clash detected, record the point
				double[] temp = StructureMath.getPointOnVector(center, v, R);
				double[] point = new double[3];
				point[0] = temp[0];
				point[1] = temp[1];
				point[2] = temp[2];
				pointList.add(point);
			}
			
			if (pointList.size() > 0){
				counter++;
				points.put(a, pointList);
			}
			
			if (atoms.size() > 100){
				if ( i % onePercent == 0 ){
					percentDone = (float)i/(float)total;
// 					Status.progress( percentDone, "Creating surface...");
					Status.progress( percentDone, M_Creatingsurface);
				}
			}
			
		}
		
		//go over the pairs to generate toroid reentrant surface patches
		double[] midpoint = new double[3];
		double[] dir = new double[3];

		//midpoint vectors are compute as rotation around z axis: for each atom pair the vectors will need to be transformed
		Vector midpointVectors = computeMidpointVectors(spacing);
		
		int z = 0;
		
//		System.out.println("pairs in = " + pairs.size());
		
		for (int i = pairs.size()-1; i >= 0; i--){
			AtomPair pair = (AtomPair)pairs.get(i);
			
			Atom a0 = pair.atoms[0];
			Atom a1 = pair.atoms[1];
			
			if (!atoms.contains(a0) && !atoms.contains(a1)){
				pairs.remove(i);
				continue;
			}
			
			Vector neighbors = (Vector)atomMap.get(a0);
			
			Vector pointList0 = (Vector)points.get(a0);
			if (pointList0 == null){
				pointList0 = new Vector();
			}
			Vector pointList1 = (Vector)points.get(a1);
			if (pointList1 == null){
				pointList1 = new Vector();
			}
			
			double a0r = ElementProperties.getAtomRadius(a0.element);
			double a1r = ElementProperties.getAtomRadius(a1.element);
			double cutoff = a0r + a1r + probeRadius*2;
			
			double dist = Algebra.distance(a0.coordinate, a1.coordinate);
			if ( dist > cutoff){
				pairs.remove(i);
				continue;
			}
			
			double half = dist/2;
			
//			System.out.println("pair = " + a0.name + ":" + a1.name);

			dir[0] = a1.coordinate[0] - a0.coordinate[0];
			dir[1] = a1.coordinate[1] - a0.coordinate[1];
			dir[2] = a1.coordinate[2] - a0.coordinate[2];
			
			Algebra.normalizeVector(dir);

			double a0probe = a0r + probeRadius;
			double a1probe = a1r + probeRadius;
			
			//determine the position of the probe
			double offA = (dist*dist + a0probe*a0probe - a1probe*a1probe)/(2*dist);
			midpoint = StructureMath.getPointOnVector(a0.coordinate, dir, offA);
			
			pair.torusCenter = midpoint;
			
			//now generate normal vectors to the bond in order to get the position of the probe center
			Vector pointers = StructureMath.transformVectors(new double[]{0,0,1}, dir, midpointVectors);
			
//			System.out.println("dir = " + dir[0] + "," + dir[1] + "," + dir[2]);
			
			if (offA > a0probe) continue;//the atoms are too far
			double offset = Math.sqrt(a0probe*a0probe - offA*offA);
			
			//check whether the probe is almost falling through the gap - if the distance between the probe center and
			//the atom-atom bond is less than probe radius, a self-intersecting surface will form
			if (offset < probeRadius){
//				System.out.println("self-intersection");
//				continue;
			}
			
			pair.torusRadius = offset;

			boolean success = false;//whether toroid has been generated or not
			
			for (int j = 0; j < pointers.size(); j++){
				double[] arrow = (double[])pointers.get(j);
				
				//the current vector points toward the center of the probe
				double[] probe = StructureMath.getPointOnVector(midpoint, arrow, offset);
				
				boolean clash = false;
				
				//now check for clashes with neighbors
				for (int k = 0; k < neighbors.size(); k++){
					Atom aa = (Atom)neighbors.get(k);
					if (aa == a0 || aa == a1) continue;
					
					if (Algebra.distance(probe, aa.coordinate) < (probeRadius + ElementProperties.getAtomRadius(aa.element))){
						clash = true;
						break;
					}
				}
				
				if (clash){
					continue;
				}
				
				//no clashes - record points that are located on the arc that connects the two atoms in the pair
				//walk along the interatom line placing points on it, getting vectors from the probe center to these points
				//and getting the points where the vectors cross the probe boundary
//				System.out.println("no clash");
//				double[] test = new double[]{ probe[0], probe[1], probe[2], -arrow[0], -arrow[1], -arrow[2] };
//				pointList.add(test);
				
				
				double ratio = offset/probeRadius;
				
				double currentDistance = 0;
				while (true){
					currentDistance += spacing*ratio;
					if (currentDistance >= dist){
//						System.out.println("break at " + a0.name + " - " + a1.name);
						break;
					}
					double[] pp = StructureMath.getPointOnVector(a0.coordinate, dir, currentDistance);
					double[] cv = Algebra.getVector(probe, pp);
					double[] out = new double[3];
					double[] temp = StructureMath.getPointOnVector(probe, cv, probeRadius);
					out[0] = temp[0];
					out[1] = temp[1];
					out[2] = temp[2];
//					out[3] = arrow[0];
//					out[4] = arrow[1];
//					out[5] = arrow[2];
					
					if (currentDistance < half){
						if (atoms.contains(a0)) pointList0.add(out);
					}
					else{
						if (atoms.contains(a1)) pointList1.add(out);
					}
					
				}
				
				success = true;

			}
			
			if (!success){
				pairs.remove(i);
			}
			
			if (atoms.contains(a0)){
				if (pointList0.size() > 0){
					points.put(a0, pointList0);
				}
			}
			
			if (atoms.contains(a1)){
				if (pointList1.size() > 0){
					points.put(a1, pointList1);
				}
			}
		
			if (atoms.size() > 100){
				if ( (atoms.size() + i) % onePercent == 0 ){
					percentDone = (float)(i + atoms.size())/(float)total;
// 					Status.progress( percentDone, "Creating surface...");
					Status.progress( percentDone, M_Creatingsurface);
				}
			}

		}
		
//		System.out.println("pairs out = " + pairs.size() + ", clashes = " + z);
		
		//look into the concave triples
		AtomGraph atomGraph = new AtomGraph(pairs, 10);
		Vector triples = atomGraph.atomTriangles;
		
//		System.err.println("triples = " + triples.size());
		
		total = atoms.size() + pairs.size() + triples.size();
		
		//walk over all triples and calculate position of the probe both above and below their plane
		//in many cases, surface will be on both sides
		for (int i = 0; i < triples.size(); i++){
			
			AtomTriangle at = (AtomTriangle)triples.get(i);
			
			//get the three atoms involved
			Atom a1 = null;
			Atom a2 = null;
			Atom a3 = null;
			
			//we only need to pairs, since they already have all three atoms
			AtomPair p1 = at.atomPairs[0];
			AtomPair p2 = at.atomPairs[1];
			AtomPair p3 = at.atomPairs[2];
			
			a1 = p1.atoms[0];
			a2 = p1.atoms[1];
			
			double[] Tik = null;
			
			//now look for a3
			if (p2.atoms[0] == a1){
				Tik = p2.torusCenter;
				a3 = p2.atoms[1];
			}
			else if (p2.atoms[0] == a2){
				a3 = p2.atoms[1];
			}
			else if (p2.atoms[1] == a1){
				Tik = p2.torusCenter;
				a3 = p2.atoms[0];
			}
			else if (p2.atoms[1] == a2){
				a3 = p2.atoms[0];
			}
			
			if (Tik == null){
				Tik = p3.torusCenter;
			}
			
			//we have three atoms, now compute the base point for the probe
			double[] Uij = Algebra.getNormalizedVector(Algebra.getVector(a1.coordinate, a2.coordinate));
			double[] Uik = Algebra.getNormalizedVector(Algebra.getVector(a1.coordinate, a3.coordinate));
			double[] Tij = p1.torusCenter;
			
			double Oijk = Math.acos(Algebra.dotProduct(Uij, Uik));//base angle
			double[] Uijk = Algebra.getNormalizedVector(Algebra.crossProduct(Uij, Uik));//base plane normal vector
			
			double[] Utb = Algebra.crossProduct(Uijk, Uij);//torus-base point normal vector
			double Rij = p1.torusRadius;
			
			Algebra.scalarMultiply(Utb, Algebra.dotProduct(Uik, Algebra.getVectorDifference(Tik, Tij)));
			Algebra.scalarMultiply(Utb, 1/Math.sin(Oijk));
			double[] Bijk = Algebra.getVectorSum(Tij, Utb);
			
			
			double r1 = ElementProperties.getAtomRadius(a1.element);
			double temp = Algebra.vectorLength(Algebra.getVectorDifference(Bijk, a1.coordinate));
			
			double pH = Math.sqrt((r1 + probeRadius)*(r1 + probeRadius) - temp*temp);
			
			Algebra.scalarMultiply(Uijk, pH);
			double[] position1 = Algebra.getVectorSum(Bijk, Uijk);
			double[] position2 = Algebra.getVectorDifference(Bijk, Uijk);
			
			//check for clashes
			Vector neighbors = (Vector)atomMap.get(a1);
			for (int j = 0; j < neighbors.size(); j++){
				Atom a = (Atom)neighbors.get(j);
				if (a == a1 || a == a2 || a == a3) continue;
				double ar = probeRadius + ElementProperties.getAtomRadius(a.element);
				double d = Algebra.distance(position1, a.coordinate);
				if ( d < ar){
					position1 = null;
					break;
				}
			}
			
			for (int j = 0; j < neighbors.size(); j++){
				Atom a = (Atom)neighbors.get(j);
				if (a == a1 || a == a2 || a == a3) continue;
				double ar = probeRadius + ElementProperties.getAtomRadius(a.element);
				double d = Algebra.distance(position2, a.coordinate);
				if ( d < ar){
					position2 = null;
					break;
				}
			}
			
			if (position1 == null && position2 == null) continue;

			//now generate points given these two positions
			
			//use the longest side of the triangle as the base to generate grid of points
			AtomPair maxPair = p1;
			double max = Algebra.distance(p1.atoms[0].coordinate, p1.atoms[1].coordinate);
			
			double v = Algebra.distance(p2.atoms[0].coordinate, p2.atoms[1].coordinate);
			if ( v > max){
				max = v;
				maxPair = p2;
			}
			
			v = Algebra.distance(p3.atoms[0].coordinate, p3.atoms[1].coordinate);
			if ( v > max){
				max = v;
				maxPair = p3;
			}
			
			//find the opposing vertex
			Atom vertex = null;
			if (maxPair == p1){
				//use p2
				Atom pa1 = p2.atoms[0];
				Atom pa2 = p2.atoms[1];
				if (maxPair.atoms[0] == pa1 || maxPair.atoms[1] == pa1){
					vertex = pa2;
				}
				else{
					vertex = pa1;
				}
			}
			else if (maxPair == p2){
				//use p2
				Atom pa1 = p1.atoms[0];
				Atom pa2 = p1.atoms[1];
				if (maxPair.atoms[0] == pa1 || maxPair.atoms[1] == pa1){
					vertex = pa2;
				}
				else{
					vertex = pa1;
				}
			}
			else{
				//use p2
				Atom pa1 = p1.atoms[0];
				Atom pa2 = p1.atoms[1];
				if (maxPair.atoms[0] == pa1 || maxPair.atoms[1] == pa1){
					vertex = pa2;
				}
				else{
					vertex = pa1;
				}
			}
			
//			System.out.println("a1 = " + maxPair.atoms[0].name + ", " + maxPair.atoms[0].coordinate[0] + "," + maxPair.atoms[0].coordinate[1] + "," + maxPair.atoms[0].coordinate[2]);
//			System.out.println("a2 = " + maxPair.atoms[1].name + ", "  + maxPair.atoms[1].coordinate[0] + "," + maxPair.atoms[1].coordinate[1] + "," + maxPair.atoms[1].coordinate[2]);
//			System.out.println("vertex = " + vertex.name + ", "  + vertex.coordinate[0] + "," + vertex.coordinate[1] + "," + vertex.coordinate[2]);
			
			
			//take the maxPair and start generating point grids
			double[] direction = Algebra.getVector(maxPair.atoms[0].coordinate, maxPair.atoms[1].coordinate);
			double[] side = Algebra.getVector(maxPair.atoms[0].coordinate, vertex.coordinate);
			double[] start = maxPair.atoms[0].coordinate;
			
			Vector pointList1 = (Vector)points.get(maxPair.atoms[0]);
			if (pointList1 == null){
				pointList1 = new Vector();
			}
			Vector pointList2 = (Vector)points.get(maxPair.atoms[1]);
			if (pointList2 == null){
				pointList2 = new Vector();
			}
			Vector pointList3 = (Vector)points.get(vertex);
			if (pointList3 == null){
				pointList3 = new Vector();
			}

			
			//find the point positions
			double current = 0;
			double length = Algebra.distance(start, maxPair.atoms[1].coordinate);
			double sideLength = Algebra.distance(start, vertex.coordinate);
			while (true){
				current += spacing*2;
				if (current > length) break;
				
				double[] currentPoint = StructureMath.getPointOnVector(start, direction, current);
				double maxDistance = ((length - current)/length)*sideLength;//distance to the opposite side of the triangle
				double currentDistance = 0;
				while (true){
					currentDistance += spacing*2;
					if (currentDistance > maxDistance) break;
					double[] p = StructureMath.getPointOnVector(currentPoint, side, currentDistance);
					if (position1 != null){
						double[] ray = Algebra.getVector(position1, p);
						double[] location = StructureMath.getPointOnVector(position1, ray, probeRadius);
						//determine which point list this location should be added to
						double min = Algebra.distance(p, maxPair.atoms[0].coordinate);
						Atom close = maxPair.atoms[0];
						
						double mm = Algebra.distance(p, maxPair.atoms[1].coordinate);
						if (mm < min){
							min = mm;
							close = maxPair.atoms[1];
						}
						
						mm = Algebra.distance(p, vertex.coordinate);
						if (mm < min){
							min = mm;
							close = vertex;
						}
						
						if (close == maxPair.atoms[0]){
							pointList1.add(location);
						}
						else if (close == maxPair.atoms[1]){
							pointList2.add(location);
						}
						else{
							pointList3.add(location);
						}
					}
					if (position2 != null){
						double[] ray = Algebra.getVector(position2, p);
						double[] location = StructureMath.getPointOnVector(position2, ray, probeRadius);
						
						//determine which point list this location should be added to
						double min = Algebra.distance(p, maxPair.atoms[0].coordinate);
						Atom close = maxPair.atoms[0];
						
						double mm = Algebra.distance(p, maxPair.atoms[1].coordinate);
						if (mm < min){
							min = mm;
							close = maxPair.atoms[1];
						}
						
						mm = Algebra.distance(p, vertex.coordinate);
						if (mm < min){
							min = mm;
							close = vertex;
						}
						
						if (close == maxPair.atoms[0]){
							pointList1.add(location);
						}
						else if (close == maxPair.atoms[1]){
							pointList2.add(location);
						}
						else{
							pointList3.add(location);
						}
					}
					
				}
			}
			
			if (pointList1.size() > 0){
				points.put(maxPair.atoms[0], pointList1);
			}
			if (pointList2.size() > 0){
				points.put(maxPair.atoms[1], pointList2);
			}
			if (pointList3.size() > 0){
				points.put(vertex, pointList3);
			}
			
			if (atoms.size() > 100){
				if ( (atoms.size() + i) % onePercent == 0 ){
					percentDone = (float)(i + atoms.size() + pairs.size())/(float)total;
// 					Status.progress( percentDone, "Creating surface...");
					Status.progress( percentDone, M_Creatingsurface);
				}
			}


		}
		
//		System.err.println("done: triples = " + triples.size());

//		if (atoms.size() > 100){
			Status.progress( 0.0f, null);
//		}

	}
	
	public HashMap getPoints(){
		return points;
	}
	
	private Vector computeVectors(double radius, double spacing){
		//let's presume that the sphere is in the origin
		//the poles
		Vector vectors = new Vector();
		
		vectors.add(new double[]{ 0,0,1 });
		vectors.add(new double[]{ 0,0,-1});
		
		//start detecting latitudes from the north pole
		//we assume sphere radius to be 1.7 A
		double currentLatitude = 0;//in degrees
		double arcOffset = 0;
		while (true){
			double offset = arcOffset + spacing;
			arcOffset = offset;
			
			//find the angle that corresponds to this offset
			double angle = offset/radius;//in radians
			
			currentLatitude = Math.toDegrees(angle);
			if (currentLatitude >= 180) break;
			
			//radius of the latitude circle
			double latR = radius*Math.sin(angle);
			double latZ = radius*Math.cos(angle);
			
			//same operation along the latitude until the circle has been completed
			//start from the x direction
			double circleOffset = 0;
			double currentAngle = 0;
			while (true){
				double arc = circleOffset + spacing;
				circleOffset = arc;
				
				//find the angle
				double a = circleOffset/latR;
				
				currentAngle = Math.toDegrees(a);
				if (currentAngle >= 360) break;
				
				double[] v = new double[]{ latR*Math.cos(a), latR*Math.sin(a), latZ };
				
				vectors.add(v);
				
			}
			
		}
		
		return vectors;
		
	}
	
	/**
	 * Returns a set of vectors used to determine positions of the probe between two atoms to
	 * define reentrant surface
	 * @param spacing
	 * @return
	 */
	private Vector computeMidpointVectors(double spacing){
		//let's presume that the sphere is in the origin
		//the poles
		Vector vectors = new Vector();
		
		double radius = 1.0;
		double z = 0.0;
		
		//start from the x direction
		double circleOffset = 0;
		double currentAngle = 0;
		while (true){
			double arc = circleOffset + spacing;
			circleOffset = arc;
			
			//find the angle
			double a = circleOffset/radius;
			
			currentAngle = Math.toDegrees(a);
			if (currentAngle > 360) break;
			
			double[] v = new double[]{ radius*Math.cos(a), radius*Math.sin(a), z };
			
			vectors.add(v);
				
		}
		
		return vectors;
		
	}

	
	public Vector getAtoms(){
		return atoms;
	}
	
	public boolean isVisible(){
		return visible;
	}
	
	public void copy(StructureComponent component){}
	
	private static String className = null;
	public static String getClassName()
	{
		if ( className == null )
			className = ((new Throwable()).getStackTrace())[0].getClassName();
		return className;
	}

	
	public String getStructureComponentType(){
		return className;
	}
	
}
